# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/Dev_Notes/51_nanogpt02.ipynb.

# %% auto 0
__all__ = ['model', 'GPTConfig', 'GPT', 'Block', 'MLP', 'CausalSelfAttention']

# %% ../nbs/Dev_Notes/51_nanogpt02.ipynb 11
import math
from dataclasses import dataclass
import torch
import torch.nn as nn
from torch.nn import functional as F
import numpy as np

# %% ../nbs/Dev_Notes/51_nanogpt02.ipynb 13
@dataclass
class GPTConfig:
    """ configurations of GPT model """
    block_size: int = 1024 # max sequence length
    vocab_size: int = 50257 # number of tokens: 50,000 BPE merges + 256 bytes tokens + 1 <|endoftext|> token
    n_layer: int = 12 # number of layers
    n_head: int = 12 # number of heads
    n_embd: int = 768 # embedding dimension

# %% ../nbs/Dev_Notes/51_nanogpt02.ipynb 17
class GPT(nn.Module):
    """  the GPT language model, with block_size max sequence length, vocab_size tokens, n_layer layers, n_head heads, n_embd embedding dimension """
    def __init__(self, config):
        super().__init__()
        self.config = config

        self.transformer = nn.ModuleDict(dict(
            # token embedding and positional embedding
            wte = nn.Embedding(config.vocab_size, config.n_embd),
            wpe = nn.Embedding(config.block_size, config.n_embd),
            # transformer: n_layer layers, each has n_head heads and n_embd embedding dimension 
            h = nn.ModuleList([Block(config) for _ in range(config.n_layer)]),
            # layernorm after the transformer blocks, added by the GPT-2 authors
            ln_f = nn.LayerNorm(config.n_embd),
        ))
        # last linear layer
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size, bias=False)

    @classmethod
    def from_pretrained(cls, model_type):
        """
        Loads pretrained GPT-2 model weights from huggingface
        from_pretrained('gpt2') loads the 124M parameter GPT-2 base model, is a constructor method that returns the GPT-2-124M model
        given the model_type, it returns the corresponding GPT model with the same architecture and weights
        """
        assert model_type in {'gpt2', 'gpt2-medium', 'gpt2-large', 'gpt2-xl'}
        from transformers import GPT2LMHeadModel
        print("loading weights from pretrained gpt: %s" % model_type)

        # n_layer, n_head and n_embd are determined from model_type
        config_args = {
            'gpt2':         dict(n_layer=12, n_head=12, n_embd=768),  # 124M params
            'gpt2-medium':  dict(n_layer=24, n_head=16, n_embd=1024), # 350M params
            'gpt2-large':   dict(n_layer=36, n_head=20, n_embd=1280), # 774M params
            'gpt2-xl':      dict(n_layer=48, n_head=25, n_embd=1600), # 1558M params
        }[model_type]
        config_args['vocab_size'] = 50257 # always 50257 for GPT model checkpoints
        config_args['block_size'] = 1024 # always 1024 for GPT model checkpoints
        # create a from-scratch initialized minGPT model
        config = GPTConfig(**config_args)
        model = GPT(config)
        sd = model.state_dict()
        sd_keys = sd.keys()
        sd_keys = [k for k in sd_keys if not k.endswith('.attn.bias')] # discard this mask / buffer, not a param

        # init a huggingface/transformers model
        model_hf = GPT2LMHeadModel.from_pretrained(model_type)
        sd_hf = model_hf.state_dict()

        # copy while ensuring all of the parameters are aligned and match in names and shapes
        sd_keys_hf = sd_hf.keys()
        sd_keys_hf = [k for k in sd_keys_hf if not k.endswith('.attn.masked_bias')] # ignore these, just a buffer
        sd_keys_hf = [k for k in sd_keys_hf if not k.endswith('.attn.bias')] # same, just the mask (buffer)
        # since the original gpt2 model was written in tensorflow so the weights are transposed - we need to transpose them back
        transposed = ['attn.c_attn.weight', 'attn.c_proj.weight', 'mlp.c_fc.weight', 'mlp.c_proj.weight']
        # basically the openai checkpoints use a "Conv1D" module, but we only want to use a vanilla Linear
        # this means that we have to transpose these weights when we import them
        assert len(sd_keys_hf) == len(sd_keys), f"mismatched keys: {len(sd_keys_hf)} != {len(sd_keys)}"
        for k in sd_keys_hf:
            if any(k.endswith(w) for w in transposed):
                # special treatment for the Conv1D weights we need to transpose
                assert sd_hf[k].shape[::-1] == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[k].t())
            else:
                # vanilla copy over the other parameters
                assert sd_hf[k].shape == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[k])

        return model

# %% ../nbs/Dev_Notes/51_nanogpt02.ipynb 21
class Block(nn.Module):
    """ Transformer Block """
    def __init__(self, config):
        super().__init__()
        self.ln_1 = nn.LayerNorm(config.n_embd)
        self.attn = CausalSelfAttention(config)
        self.ln_2 = nn.LayerNorm(config.n_embd)
        self.mlp = MLP(config)

    def forward(self, x):
        # skip connection around attention and feed forward
        # different from the original transformer, layer normalization is applied inside/before the residual connections
        # since clean residual pathways are desirable from optimization perspective
        x = x + self.attn(self.ln_1(x)) # reduced operation
        x = x + self.mlp(self.ln_2(x))  # mapped operation 
        return x

# %% ../nbs/Dev_Notes/51_nanogpt02.ipynb 25
class MLP(nn.Module):
    """ Multi-Layer Perceptron """
    def __init__(self, config):
        super().__init__()
        self.c_fc    = nn.Linear(config.n_embd, 4 * config.n_embd)
        self.gelu    = nn.GELU(approximate='tanh') # historic reason for the tanh-approximation
        self.c_proj  = nn.Linear(4 * config.n_embd, config.n_embd)

    def forward(self, x):
        # two linear transformations with a GELU activation in between
        x = self.c_fc(x)
        x = self.gelu(x)
        x = self.c_proj(x)
        return x

# %% ../nbs/Dev_Notes/51_nanogpt02.ipynb 29
class CausalSelfAttention(nn.Module):
    """ Causal Multi-Head Self-Attention Module """
    def __init__(self, config):
        super().__init__()
        assert config.n_embd % config.n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(config.n_embd, 3 * config.n_embd)
        # output projection
        self.c_proj = nn.Linear(config.n_embd, config.n_embd)
        # regularization
        self.n_head = config.n_head
        self.n_embd = config.n_embd
        # not really a 'bias', more of a mask, but following the OpenAI/HF naming though
        self.register_buffer("bias", torch.tril(torch.ones(config.block_size, config.block_size))
                                     .view(1, 1, config.block_size, config.block_size))

    def forward(self, x):
        B, T, C = x.size() # batch size, sequence length, embedding dimensionality (n_embd)
        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        # nh is "number of heads", hs is "head size", and C (number of channels) = nh * hs
        # e.g. in GPT-2 (124M), n_head=12, hs=64, so nh*hs=C=768 channels in the Transformer
        qkv = self.c_attn(x)
        q, k, v = qkv.split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        # attention (materializes the large (T,T) matrix for all the queries and keys)
        att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        # auto-regressive mask (upper triangular) - make sure the attention to only the left in the input sequence i.e. past never to the future
        att = att.masked_fill(self.bias[:,:,:T,:T] == 0, float('-inf'))
        # normalize the attentions
        att = F.softmax(att, dim=-1)
        # matrix multiplication i.e. weighted sum of the values
        y = att @ v # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        # this actually performs concatination of all heads
        y = y.transpose(1, 2).contiguous().view(B, T, C) # re-assemble all head outputs side by side
        # output projection
        y = self.c_proj(y)
        return y

# %% ../nbs/Dev_Notes/51_nanogpt02.ipynb 32
#|eval: false
model = GPT.from_pretrained('gpt2')
print("Successfully load the GPT-2-124M model with pre-trained weights from Huggingface!!!")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/Dev_Notes/51_nanogpt09.ipynb.

# %% auto 0
__all__ = ['ddp', 'total_batch_size', 'B', 'T', 'grad_accum_steps', 'train_loader', 'val_loader', 'model', 'use_compile',
           'raw_model', 'enc', 'log_dir', 'log_file', 'optimizer', 'losses', 'norms', 'lrs', 'GPTConfig', 'GPT',
           'Block', 'MLP', 'CausalSelfAttention', 'load_tokens', 'DataLoaderLite']

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 11
import math
from dataclasses import dataclass
import torch
import torch.nn as nn
from torch.nn import functional as F
import numpy as np
import tiktoken
import matplotlib.pyplot as plt
import time
import inspect
import io
# from hellaswag import render_example, iterate_examples

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 13
@dataclass
class GPTConfig:
    """ configurations of GPT model """
    block_size: int = 1024 # max sequence length
    vocab_size: int = 50257 # number of tokens: 50,000 BPE merges + 256 bytes tokens + 1 <|endoftext|> token
    n_layer: int = 12 # number of layers
    n_head: int = 12 # number of heads
    n_embd: int = 768 # embedding dimension

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 19
class GPT(nn.Module):
    """  the GPT language model, with block_size max sequence length, vocab_size tokens, n_layer layers, n_head heads, n_embd embedding dimension """
    def __init__(self, config):
        super().__init__()
        self.config = config

        self.transformer = nn.ModuleDict(dict(
            # token embedding and positional embedding
            wte = nn.Embedding(config.vocab_size, config.n_embd),
            wpe = nn.Embedding(config.block_size, config.n_embd),
            # transformer: n_layer layers, each has n_head heads and n_embd embedding dimension 
            h = nn.ModuleList([Block(config) for _ in range(config.n_layer)]),
            # layernorm after the transformer blocks, added by the GPT-2 authors
            ln_f = nn.LayerNorm(config.n_embd),
        ))
        # last linear layer
        self.lm_head = nn.Linear(config.n_embd, config.vocab_size, bias=False)

        # weight sharing scheme - as we noted when visualizing the wights in the first notebook
        self.transformer.wte.weight = self.lm_head.weight

        # init params
        self.apply(self._init_weights)
    
    def _init_weights(self, module):
        """ initialize weights - normal distribution with 0 mean and 0.02 std, 0.02 ~ 1/sqrt(#params) """
        if isinstance(module, nn.Linear):
            std = 0.02
        if hasattr(module, 'NANOGPT_SCALE_INIT'):
            std *= (2 * self.config.n_layer) ** -0.5 # 2 becuase of MLP and Attention
            torch.nn.init.normal_(module.weight, mean=0.0, std=std)
            if module.bias is not None:
                torch.nn.init.zeros_(module.bias) # this is not pytorch default, but was done in the original code
        elif isinstance(module, nn.Embedding):
            torch.nn.init.normal_(module.weight, mean=0.0, std=0.02) 

    def forward(self, idx, targets=None):
        # idx are tokens of shape (B, T), where B is batch size, T is the length of the sequence, T<=block_size
        B, T = idx.size()
        assert T <= self.config.block_size, f"Cannot forward sequence of length {T}, block size is only {self.config.block_size}"
        # forward the token and posisition embeddings
        pos = torch.arange(0, T, dtype=torch.long, device=idx.device) # shape (T)
        pos_emb = self.transformer.wpe(pos) # position embeddings of shape (T, n_embd)
        tok_emb = self.transformer.wte(idx) # token embeddings of shape (B, T, n_embd)
        # broadcast along batch dimension, since the same position embeddings are used for all item in a batch
        x = tok_emb + pos_emb
        # forward the blocks of the transformer
        for block in self.transformer.h:
            x = block(x)
        # forward the final layernorm and the classifier
        x = self.transformer.ln_f(x)
        logits = self.lm_head(x) # (B, T, vocab_size)
        # adding cross-entropy loss calculation
        loss = None
        if targets is not None:
            loss = F.cross_entropy(logits.view(-1, logits.size(-1)), targets.view(-1))
        return logits, loss


    @classmethod
    def from_pretrained(cls, model_type):
        """
        Loads pretrained GPT-2 model weights from huggingface
        from_pretrained('gpt2') loads the 124M parameter GPT-2 base model, is a constructor method that returns the GPT-2-124M model
        given the model_type, it returns the corresponding GPT model with the same architecture and weights
        """
        assert model_type in {'gpt2', 'gpt2-medium', 'gpt2-large', 'gpt2-xl'}
        from transformers import GPT2LMHeadModel
        print("loading weights from pretrained gpt: %s" % model_type)

        # n_layer, n_head and n_embd are determined from model_type
        config_args = {
            'gpt2':         dict(n_layer=12, n_head=12, n_embd=768),  # 124M params
            'gpt2-medium':  dict(n_layer=24, n_head=16, n_embd=1024), # 350M params
            'gpt2-large':   dict(n_layer=36, n_head=20, n_embd=1280), # 774M params
            'gpt2-xl':      dict(n_layer=48, n_head=25, n_embd=1600), # 1558M params
        }[model_type]
        config_args['vocab_size'] = 50257 # always 50257 for GPT model checkpoints
        config_args['block_size'] = 1024 # always 1024 for GPT model checkpoints
        # create a from-scratch initialized minGPT model
        config = GPTConfig(**config_args)
        model = GPT(config)
        sd = model.state_dict()
        sd_keys = sd.keys()
        sd_keys = [k for k in sd_keys if not k.endswith('.attn.bias')] # discard this mask / buffer, not a param

        # init a huggingface/transformers model
        model_hf = GPT2LMHeadModel.from_pretrained(model_type)
        sd_hf = model_hf.state_dict()

        # copy while ensuring all of the parameters are aligned and match in names and shapes
        sd_keys_hf = sd_hf.keys()
        sd_keys_hf = [k for k in sd_keys_hf if not k.endswith('.attn.masked_bias')] # ignore these, just a buffer
        sd_keys_hf = [k for k in sd_keys_hf if not k.endswith('.attn.bias')] # same, just the mask (buffer)
        # since the original gpt2 model was written in tensorflow so the weights are transposed - we need to transpose them back
        transposed = ['attn.c_attn.weight', 'attn.c_proj.weight', 'mlp.c_fc.weight', 'mlp.c_proj.weight']
        # basically the openai checkpoints use a "Conv1D" module, but we only want to use a vanilla Linear
        # this means that we have to transpose these weights when we import them
        assert len(sd_keys_hf) == len(sd_keys), f"mismatched keys: {len(sd_keys_hf)} != {len(sd_keys)}"
        for k in sd_keys_hf:
            if any(k.endswith(w) for w in transposed):
                # special treatment for the Conv1D weights we need to transpose
                assert sd_hf[k].shape[::-1] == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[k].t())
            else:
                # vanilla copy over the other parameters
                assert sd_hf[k].shape == sd[k].shape
                with torch.no_grad():
                    sd[k].copy_(sd_hf[k])

        return model
    
    # adding weight decay configuration
    def configure_optimizers(self, weight_decay, learning_rate, device):
        # start with all of the candidate parameters (that require grad)
        param_dict = {pn: p for pn, p in self.named_parameters()}
        param_dict = {pn: p for pn, p in param_dict.items() if p.requires_grad}
        # create optim groups. Any parameters that is 2D will be weight decayed, otherwise no.
        # i.e. all weight tensors in matmuls + embeddings decay, all biases and layernorms don't.
        decay_params = [p for n, p in param_dict.items() if p.dim() >= 2]
        nodecay_params = [p for n, p in param_dict.items() if p.dim() < 2]
        optim_groups = [
            {'params': decay_params, 'weight_decay': weight_decay},
            {'params': nodecay_params, 'weight_decay': 0.0}
        ]
        num_decay_params = sum(p.numel() for p in decay_params)
        num_nodecay_params = sum(p.numel() for p in nodecay_params)
        # print(f"num decayed parameter tensors: {len(decay_params)}, with {num_decay_params:,} parameters")
        # print(f"num non-decayed parameter tensors: {len(nodecay_params)}, with {num_nodecay_params:,} parameters")
        if master_process:
            print(f"num decayed parameter tensors: {len(decay_params)}, with {num_decay_params:,} parameters")
            print(f"num non-decayed parameter tensors: {len(nodecay_params)}, with {num_nodecay_params:,} parameters")
        # Create AdamW optimizer and use the fused version if it is available
        fused_available = 'fused' in inspect.signature(torch.optim.AdamW).parameters
        use_fused = fused_available and 'cuda' in device
        # print(f"using fused AdamW: {use_fused}")
        if master_process:
            print(f"using fused AdamW: {use_fused}")
        optimizer = torch.optim.AdamW(optim_groups, lr=learning_rate, betas=(0.9, 0.95), eps=1e-8, fused=use_fused)
        return optimizer

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 23
class Block(nn.Module):
    """ Transformer Block """
    def __init__(self, config):
        super().__init__()
        self.ln_1 = nn.LayerNorm(config.n_embd)
        self.attn = CausalSelfAttention(config)
        self.ln_2 = nn.LayerNorm(config.n_embd)
        self.mlp = MLP(config)

    def forward(self, x):
        # skip connection around attention and feed forward
        # different from the original transformer, layer normalization is applied inside/before the residual connections
        # since clean residual pathways are desirable from optimization perspective
        x = x + self.attn(self.ln_1(x)) # reduced operation
        x = x + self.mlp(self.ln_2(x))  # mapped operation 
        return x

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 27
class MLP(nn.Module):
    """ Multi-Layer Perceptron """
    def __init__(self, config):
        super().__init__()
        self.c_fc    = nn.Linear(config.n_embd, 4 * config.n_embd)
        self.gelu    = nn.GELU(approximate='tanh') # historic reason for the tanh-approximation
        self.c_proj  = nn.Linear(4 * config.n_embd, config.n_embd)
        # special flag
        self.c_proj.NANOGPT_SCALE_INIT = 1

    def forward(self, x):
        # two linear transformations with a GELU activation in between
        x = self.c_fc(x)
        x = self.gelu(x)
        x = self.c_proj(x)
        return x

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 31
class CausalSelfAttention(nn.Module):
    """ Causal Multi-Head Self-Attention Module """
    def __init__(self, config):
        super().__init__()
        assert config.n_embd % config.n_head == 0
        # key, query, value projections for all heads, but in a batch
        self.c_attn = nn.Linear(config.n_embd, 3 * config.n_embd)
        # output projection
        self.c_proj = nn.Linear(config.n_embd, config.n_embd)
        # special flag
        self.c_proj.NANOGPT_SCALE_INIT = 1
        # regularization
        self.n_head = config.n_head
        self.n_embd = config.n_embd
        # not really a 'bias', more of a mask, but following the OpenAI/HF naming though
        self.register_buffer("bias", torch.tril(torch.ones(config.block_size, config.block_size))
                                     .view(1, 1, config.block_size, config.block_size))

    def forward(self, x):
        B, T, C = x.size() # batch size, sequence length, embedding dimensionality (n_embd)
        # calculate query, key, values for all heads in batch and move head forward to be the batch dim
        # nh is "number of heads", hs is "head size", and C (number of channels) = nh * hs
        # e.g. in GPT-2 (124M), n_head=12, hs=64, so nh*hs=C=768 channels in the Transformer
        qkv = self.c_attn(x)
        q, k, v = qkv.split(self.n_embd, dim=2)
        k = k.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        q = q.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        v = v.view(B, T, self.n_head, C // self.n_head).transpose(1, 2) # (B, nh, T, hs)
        
        # attention (materializes the large (T,T) matrix for all the queries and keys)
        # att = (q @ k.transpose(-2, -1)) * (1.0 / math.sqrt(k.size(-1)))
        # auto-regressive mask (upper triangular) - make sure the attention to only the left in the input sequence i.e. past never to the future
        # att = att.masked_fill(self.bias[:,:,:T,:T] == 0, float('-inf'))
        # normalize the attentions
        # att = F.softmax(att, dim=-1)
        # matrix multiplication i.e. weighted sum of the values
        # y = att @ v # (B, nh, T, T) x (B, nh, T, hs) -> (B, nh, T, hs)
        # this actually performs concatination of all heads
        
        # FlashAttention to replace the above 4 lines of code
        y = F.scaled_dot_product_attention(q, k, v, is_causal=True) # flash attention
        
        y = y.transpose(1, 2).contiguous().view(B, T, C) # re-assemble all head outputs side by side
        # output projection
        y = self.c_proj(y)
        return y

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 34
def load_tokens(filename):
    """ load tokens from file """
    npt = np.load(filename)
    ptt = torch.tensor(npt, dtype=torch.long)
    return ptt

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 35
class DataLoaderLite():
    """ Simple Dataloader for loading tokens from `input.txt` save on disk and serving them as batches"""
    def __init__(self, B, T, process_rank, num_processes, split):
        self.B = B
        self.T = T
        self.process_rank = process_rank
        self.num_processes = num_processes

        assert split in {'train', 'val'}

        # get the shard filenames
        data_root = "edu_fineweb10B"
        shards = os.listdir(data_root)
        shards = [s for s in shards if split in s]
        shards = sorted(shards)
        shards = [os.path.join(data_root, s) for s in shards]
        self.shards = shards
        assert len(shards) > 0, f"no shards found for split {split}"

        if master_process:
            print(f"found {len(shards)} shards for split {split}")
        self.reset()

    def reset(self):
        # state, init at shard zero
        self.current_shard = 0
        self.tokens = load_tokens(self.shards[self.current_shard])


        # state, init at shard zero
        self.current_shard = 0
        self.tokens = load_tokens(self.shards[self.current_shard])
        self.current_position = self.B * self.T * self.process_rank

    def next_batch(self):
        B, T = self.B, self.T
        buf = self.tokens[self.current_position : self.current_position+B*T+1]
        x = (buf[:-1]).view(B, T) # inputs
        y = (buf[1:]).view(B, T) # targets
        # advance the position in the tensor
        # self.current_position += B * T
        self.current_position += B * T * self.num_processes
        # if loading the next batch would be out of bounds, advance to next shard
        if self.current_position + (B * T * self.num_processes + 1) > len(self.tokens):
            self.current_shard = (self.current_shard + 1) % len(self.shards)
            self.tokens = load_tokens(self.shards[self.current_shard])
            self.current_position = B * T * self.process_rank
        return x, y

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 41
#|eval: false
torch.manual_seed(1337)
if torch.cuda.is_available():
    torch.cuda.manual_seed(1337)

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 76
#|eval: false
# run the training loop
from torch.distributed import init_process_group, destroy_process_group
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.distributed as dist

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 77
#|eval: false
# simple launch:
# python train_gpt2.py
# DDP launch for e.g. 8 GPUs:
# torchrun --standalone --nproc_per_node=8 train_gpt2.py

# set up DDP (distributed data parallel).
# torchrun command sets the env variables RANK, LOCAL_RANK, and WORLD_SIZE
ddp = int(os.environ.get('RANK', -1)) != -1 # is this a ddp run?
if ddp:
    # use of DDP atm demands CUDA, we set the device appropriately according to rank
    assert torch.cuda.is_available(), "for now i think we need CUDA for DDP"
    init_process_group(backend='nccl')
    ddp_rank = int(os.environ['RANK'])
    ddp_local_rank = int(os.environ['LOCAL_RANK'])
    ddp_world_size = int(os.environ['WORLD_SIZE'])
    device = f'cuda:{ddp_local_rank}'
    torch.cuda.set_device(device)
    master_process = ddp_rank == 0 # this process will do logging, checkpointing etc.
else:
    # vanilla, non-DDP run
    ddp_rank = 0
    ddp_local_rank = 0
    ddp_world_size = 1
    master_process = True
    # attempt to autodetect device
    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda"
    elif hasattr(torch.backends, "mps") and torch.backends.mps.is_available():
        device = "mps"
    print(f"using device: {device}")

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 79
#|eval: false
# initialize the training data loader
# train_loader = DataLoaderLite(B=16, T=512)
total_batch_size = 524288 # 2**19, ~0.5M, in number of tokens
B = 16 # micro batch size
T = 1024 # sequence length

# assert total_batch_size % (B * T) == 0, "make sure total_batch_size is divisible by B * T"
# grad_accum_steps = total_batch_size // (B * T)
# print(f"total desired batch size: {total_batch_size}")
# print(f"=> calculated gradient accumulation steps: {grad_accum_steps}")

assert total_batch_size % (B * T * ddp_world_size) == 0, "make sure total_batch_size is divisible by B * T * ddp_world_size"
grad_accum_steps = total_batch_size // (B * T * ddp_world_size)
if master_process:
    print(f"total desired batch size: {total_batch_size}")
    print(f"=> calculated gradient accumulation steps: {grad_accum_steps}")


# train_loader = DataLoaderLite(B=B, T=T)

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 80
#|eval: false
train_loader = DataLoaderLite(B=B, T=T, process_rank=ddp_rank, num_processes=ddp_world_size, split='train')
val_loader = DataLoaderLite(B=B, T=T, process_rank=ddp_rank, num_processes=ddp_world_size, split="val")

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 82
#|eval: false
# the default precision is 'float32' for training
# use tf32 precision for training, that will accelerate training
torch.set_float32_matmul_precision('high')
# change vocab size from 50257 to 50304 (has a pump at 31th step), try 50264%8==0, used 50264 to avoid the bump
model = GPT(GPTConfig(vocab_size=50264))
# model.eval() # commented out since we are training
model.to(device)
# adding torch.compile. It costs compilation time but the training is faster
use_compile = True # torch.compile interferes with HellaSwag eval and Generation. TODO fix
if use_compile:
    model = torch.compile(model)

if ddp:
    model = DDP(model, device_ids=[ddp_local_rank])
raw_model = model.module if ddp else model # always contains the "raw" unwrapped model

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 91
#|eval: false
# prefix tokens
enc = tiktoken.get_encoding('gpt2')

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 92
#|eval: false
# create the log directory we will write checkpoints to and log to
log_dir = "log"
os.makedirs(log_dir, exist_ok=True)
log_file = os.path.join(log_dir, f"log.txt")
with open(log_file, "w") as f: # open for writing to clear the file
    pass

# %% ../nbs/Dev_Notes/51_nanogpt09.ipynb 93
#|eval: false
# create the optimizer object; added GPT-3's params for AdamW
# optimizer = torch.optim.AdamW(model.parameters(), lr=3e-4, betas=(0.9, 0.95), eps=1e-8)
# optimizer = model.configure_optimizers(weight_decay=0.1, learning_rate=6e-4, device=device)
optimizer = raw_model.configure_optimizers(weight_decay=0.1, learning_rate=6e-4, device=device)
losses, norms, lrs = [], [], []
for step in range(max_steps):
    t0 = time.time()
    last_step = (step == max_steps - 1)
    # once in a while evaluate our validation loss
    if step % val_check_steps == 0 or last_step:
        model.eval()
        val_loader.reset()
        with torch.no_grad():
            val_loss_accum = 0.0
            val_loss_steps = 20
            for _ in range(val_loss_steps):
                x, y = val_loader.next_batch()
                x, y = x.to(device), y.to(device)
                with torch.autocast(device_type=device, dtype=torch.bfloat16):
                    logits, loss = model(x, y)
                loss = loss / val_loss_steps
                val_loss_accum += loss.detach()
        if ddp:
            dist.all_reduce(val_loss_accum, op=dist.ReduceOp.AVG)
        if master_process:
            print(f"validation loss: {val_loss_accum.item():.4f}")
            with open(log_file, "a") as f:
                f.write(f"{step} val {val_loss_accum.item():.4f}\n")

    # once in a while evaluate hellaswag
    if (step % val_check_steps == 0 or last_step):# and (not use_compile):
        num_correct_norm = 0
        num_total = 0
        for i, example in enumerate(iterate_examples("val")):
            # only process examples where i % ddp_world_size == ddp_rank
            if i % ddp_world_size != ddp_rank:
                continue
            # render the example into tokens and labels
            _, tokens, mask, label = render_example(example)
            tokens = tokens.to(device)
            mask = mask.to(device)
            # get the logits
            with torch.no_grad():
                with torch.autocast(device_type=device, dtype=torch.bfloat16):
                    logits, loss = model(tokens)
                pred_norm = get_most_likely_row(tokens, mask, logits)
            num_total += 1
            num_correct_norm += int(pred_norm == label)
        # reduce the stats across all processes
        if ddp:
            num_total = torch.tensor(num_total, dtype=torch.long, device=device)
            num_correct_norm = torch.tensor(num_correct_norm, dtype=torch.long, device=device)
            dist.all_reduce(num_total, op=dist.ReduceOp.SUM)
            dist.all_reduce(num_correct_norm, op=dist.ReduceOp.SUM)
            num_total = num_total.item()
            num_correct_norm = num_correct_norm.item()
        acc_norm = num_correct_norm / num_total
        if master_process:
            print(f"HellaSwag accuracy: {num_correct_norm}/{num_total}={acc_norm:.4f}")
            with open(log_file, "a") as f:
                f.write(f"{step} hella {acc_norm:.4f}\n")

    # once in a while generate from the model (except step 0, which is noise)
    # disabled because torch.compile throws a scary error i can't solve rn
    # if you disable torch.compile, this code works fine
    if ((step > 0 and step % val_check_steps == 0) or last_step):# and (not use_compile):
        model.eval()
        num_return_sequences = 4
        max_length = 32
        tokens = enc.encode("Hello, I'm a language model,")
        tokens = torch.tensor(tokens, dtype=torch.long)
        tokens = tokens.unsqueeze(0).repeat(num_return_sequences, 1)
        xgen = tokens.to(device)
        sample_rng = torch.Generator(device=device)
        sample_rng.manual_seed(42 + ddp_rank)
        while xgen.size(1) < max_length:
            # forward the model to get the logits
            with torch.no_grad():
                # logits, loss = model(xgen) # (B, T, vocab_size)
                with torch.autocast(device_type=device, dtype=torch.bfloat16):
                    logits, loss = model(xgen) # (B, T, vocab_size)
                # take the logits at the last position
                logits = logits[:, -1, :] # (B, vocab_size)
                # get the probabilities
                probs = F.softmax(logits, dim=-1)
                # do top-k sampling of 50 (huggingface pipeline default)
                # topk_probs here becomes (5, 50), topk_indices is (5, 50)
                topk_probs, topk_indices = torch.topk(probs, 50, dim=-1)
                # select a token from the top-k probabilities
                # note: multinomial does not demand the input to sum to 1
                ix = torch.multinomial(topk_probs, 1, generator=sample_rng) # (B, 1)
                # gather the corresponding indices
                xcol = torch.gather(topk_indices, -1, ix) # (B, 1)
                # append to the sequence
                xgen = torch.cat((xgen, xcol), dim=1)
        # print the generated text
        for i in range(num_return_sequences):
            tokens = xgen[i, :max_length].tolist()
            decoded = enc.decode(tokens)
            print(f"rank {ddp_rank} sample {i}: {decoded}")

    # training loop
    model.train()
    loss_accum = 0.0
    for micro_step in range(grad_accum_steps):
        x, y = train_loader.next_batch()
        x, y = x.to(device), y.to(device)
        with torch.autocast(device_type=device, dtype=torch.bfloat16):
            logits, loss = model(x, y)
        # we have to scale the loss to account for gradient accumulation,
        # because the gradients just add on each successive backward().
        # addition of gradients corresponds to a SUM in the objective, but
        # instead of a SUM we want MEAN. Scale the loss here so it comes out right
        loss = loss / grad_accum_steps
        loss_accum += loss.detach()
        if ddp:
            model.require_backward_grad_sync = (micro_step == grad_accum_steps - 1)
        loss.backward()
    if ddp:
        dist.all_reduce(loss_accum, op=dist.ReduceOp.AVG)
    losses.append(loss_accum.item())
    # added gradient clipping
    norm = torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)
    norms.append(norm.item())
    # determine and set the learning rate for this iteration
    lr = get_lr(step)
    lrs.append(lr)
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr
    optimizer.step()
    torch.cuda.synchronize() # wait for the GPU to finish work
    t1 = time.time()
    dt = t1 - t0 # time difference in seconds
    # tokens_processed = train_loader.B * train_loader.T
    # tokens_processed = train_loader.B * train_loader.T * grad_accum_steps
    tokens_processed = train_loader.B * train_loader.T * grad_accum_steps * ddp_world_size
    tokens_per_sec = tokens_processed / dt
    # print(f"step {i:4d} | loss: {loss.item():.6f} | dt: {dt*1000:.2f}ms | tok/sec: {tokens_per_sec:.2f}")
    # print(f"step {i:4d} | loss: {loss.item():.6f} | norm: {norm.item():.4f} | dt: {dt*1000:.2f}ms | tok/sec: {tokens_per_sec:.2f}")
    # print(f"step {step:4d} | loss: {loss_accum.item():.6f} | lr {lr:.4e} | norm: {norm.item():.4f} | dt: {dt*1000:.2f}ms | tok/sec: {tokens_per_sec:.2f}")
    if master_process:
        print(f"step {step:5d} | loss: {loss_accum.item():.6f} | lr {lr:.4e} | norm: {norm.item():.4f} | dt: {dt*1000:.2f}ms | tok/sec: {tokens_per_sec:.2f}")
        with open(log_file, "a") as f:
            f.write(f"{step} train {loss_accum.item():.6f}\n")
            
if ddp:
    destroy_process_group()



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/buildNanoGPT',
                'doc_host': 'https://hdocmsu.github.io',
                'git_url': 'https://github.com/hdocmsu/buildNanoGPT/',
                'lib_path': 'buildNanoGPT'},
  'syms': { 'buildNanoGPT.nanogpt1': { 'buildNanoGPT.nanogpt1.Block': ('Dev_Notes/nanogpt01.html#block', 'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.Block.__init__': ( 'Dev_Notes/nanogpt01.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.Block.forward': ( 'Dev_Notes/nanogpt01.html#block.forward',
                                                                                'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.CausalSelfAttention': ( 'Dev_Notes/nanogpt01.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt01.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt01.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.GPT': ('Dev_Notes/nanogpt01.html#gpt', 'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.GPT.__init__': ( 'Dev_Notes/nanogpt01.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.GPT.forward': ( 'Dev_Notes/nanogpt01.html#gpt.forward',
                                                                              'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.GPTConfig': ( 'Dev_Notes/nanogpt01.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.MLP': ('Dev_Notes/nanogpt01.html#mlp', 'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.MLP.__init__': ( 'Dev_Notes/nanogpt01.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt1.py'),
                                       'buildNanoGPT.nanogpt1.MLP.forward': ( 'Dev_Notes/nanogpt01.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt1.py')},
            'buildNanoGPT.nanogpt10': { 'buildNanoGPT.nanogpt10.Block': ('nanoGPT_PyPI/nanogpt10.html#block', 'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.Block.__init__': ( 'nanoGPT_PyPI/nanogpt10.html#block.__init__',
                                                                                   'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.Block.forward': ( 'nanoGPT_PyPI/nanogpt10.html#block.forward',
                                                                                  'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.CausalSelfAttention': ( 'nanoGPT_PyPI/nanogpt10.html#causalselfattention',
                                                                                        'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.CausalSelfAttention.__init__': ( 'nanoGPT_PyPI/nanogpt10.html#causalselfattention.__init__',
                                                                                                 'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.CausalSelfAttention.forward': ( 'nanoGPT_PyPI/nanogpt10.html#causalselfattention.forward',
                                                                                                'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.DataLoaderLite': ( 'nanoGPT_PyPI/nanogpt10.html#dataloaderlite',
                                                                                   'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.DataLoaderLite.__init__': ( 'nanoGPT_PyPI/nanogpt10.html#dataloaderlite.__init__',
                                                                                            'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.DataLoaderLite.next_batch': ( 'nanoGPT_PyPI/nanogpt10.html#dataloaderlite.next_batch',
                                                                                              'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.DataLoaderLite.reset': ( 'nanoGPT_PyPI/nanogpt10.html#dataloaderlite.reset',
                                                                                         'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.GPT': ('nanoGPT_PyPI/nanogpt10.html#gpt', 'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.GPT.__init__': ( 'nanoGPT_PyPI/nanogpt10.html#gpt.__init__',
                                                                                 'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.GPT._init_weights': ( 'nanoGPT_PyPI/nanogpt10.html#gpt._init_weights',
                                                                                      'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.GPT.configure_optimizers': ( 'nanoGPT_PyPI/nanogpt10.html#gpt.configure_optimizers',
                                                                                             'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.GPT.forward': ( 'nanoGPT_PyPI/nanogpt10.html#gpt.forward',
                                                                                'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.GPT.from_pretrained': ( 'nanoGPT_PyPI/nanogpt10.html#gpt.from_pretrained',
                                                                                        'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.GPTConfig': ( 'nanoGPT_PyPI/nanogpt10.html#gptconfig',
                                                                              'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.MLP': ('nanoGPT_PyPI/nanogpt10.html#mlp', 'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.MLP.__init__': ( 'nanoGPT_PyPI/nanogpt10.html#mlp.__init__',
                                                                                 'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.MLP.forward': ( 'nanoGPT_PyPI/nanogpt10.html#mlp.forward',
                                                                                'buildNanoGPT/nanogpt10.py'),
                                        'buildNanoGPT.nanogpt10.load_tokens': ( 'nanoGPT_PyPI/nanogpt10.html#load_tokens',
                                                                                'buildNanoGPT/nanogpt10.py')},
            'buildNanoGPT.nanogpt2': { 'buildNanoGPT.nanogpt2.Block': ('Dev_Notes/nanogpt02.html#block', 'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.Block.__init__': ( 'Dev_Notes/nanogpt02.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.Block.forward': ( 'Dev_Notes/nanogpt02.html#block.forward',
                                                                                'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.CausalSelfAttention': ( 'Dev_Notes/nanogpt02.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt02.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt02.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.GPT': ('Dev_Notes/nanogpt02.html#gpt', 'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.GPT.__init__': ( 'Dev_Notes/nanogpt02.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.GPT.from_pretrained': ( 'Dev_Notes/nanogpt02.html#gpt.from_pretrained',
                                                                                      'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.GPTConfig': ( 'Dev_Notes/nanogpt02.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.MLP': ('Dev_Notes/nanogpt02.html#mlp', 'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.MLP.__init__': ( 'Dev_Notes/nanogpt02.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt2.py'),
                                       'buildNanoGPT.nanogpt2.MLP.forward': ( 'Dev_Notes/nanogpt02.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt2.py')},
            'buildNanoGPT.nanogpt3': { 'buildNanoGPT.nanogpt3.Block': ('Dev_Notes/nanogpt03.html#block', 'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.Block.__init__': ( 'Dev_Notes/nanogpt03.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.Block.forward': ( 'Dev_Notes/nanogpt03.html#block.forward',
                                                                                'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.CausalSelfAttention': ( 'Dev_Notes/nanogpt03.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt03.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt03.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.GPT': ('Dev_Notes/nanogpt03.html#gpt', 'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.GPT.__init__': ( 'Dev_Notes/nanogpt03.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.GPT.forward': ( 'Dev_Notes/nanogpt03.html#gpt.forward',
                                                                              'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.GPT.from_pretrained': ( 'Dev_Notes/nanogpt03.html#gpt.from_pretrained',
                                                                                      'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.GPTConfig': ( 'Dev_Notes/nanogpt03.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.MLP': ('Dev_Notes/nanogpt03.html#mlp', 'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.MLP.__init__': ( 'Dev_Notes/nanogpt03.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt3.py'),
                                       'buildNanoGPT.nanogpt3.MLP.forward': ( 'Dev_Notes/nanogpt03.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt3.py')},
            'buildNanoGPT.nanogpt4': { 'buildNanoGPT.nanogpt4.Block': ('Dev_Notes/nanogpt04.html#block', 'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.Block.__init__': ( 'Dev_Notes/nanogpt04.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.Block.forward': ( 'Dev_Notes/nanogpt04.html#block.forward',
                                                                                'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.CausalSelfAttention': ( 'Dev_Notes/nanogpt04.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt04.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt04.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.GPT': ('Dev_Notes/nanogpt04.html#gpt', 'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.GPT.__init__': ( 'Dev_Notes/nanogpt04.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.GPT.forward': ( 'Dev_Notes/nanogpt04.html#gpt.forward',
                                                                              'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.GPT.from_pretrained': ( 'Dev_Notes/nanogpt04.html#gpt.from_pretrained',
                                                                                      'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.GPTConfig': ( 'Dev_Notes/nanogpt04.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.MLP': ('Dev_Notes/nanogpt04.html#mlp', 'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.MLP.__init__': ( 'Dev_Notes/nanogpt04.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt4.py'),
                                       'buildNanoGPT.nanogpt4.MLP.forward': ( 'Dev_Notes/nanogpt04.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt4.py')},
            'buildNanoGPT.nanogpt5': { 'buildNanoGPT.nanogpt5.Block': ('Dev_Notes/nanogpt05.html#block', 'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.Block.__init__': ( 'Dev_Notes/nanogpt05.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.Block.forward': ( 'Dev_Notes/nanogpt05.html#block.forward',
                                                                                'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.CausalSelfAttention': ( 'Dev_Notes/nanogpt05.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt05.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt05.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.DataLoaderLite': ( 'Dev_Notes/nanogpt05.html#dataloaderlite',
                                                                                 'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.DataLoaderLite.__init__': ( 'Dev_Notes/nanogpt05.html#dataloaderlite.__init__',
                                                                                          'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.DataLoaderLite.next_batch': ( 'Dev_Notes/nanogpt05.html#dataloaderlite.next_batch',
                                                                                            'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.GPT': ('Dev_Notes/nanogpt05.html#gpt', 'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.GPT.__init__': ( 'Dev_Notes/nanogpt05.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.GPT._init_weights': ( 'Dev_Notes/nanogpt05.html#gpt._init_weights',
                                                                                    'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.GPT.forward': ( 'Dev_Notes/nanogpt05.html#gpt.forward',
                                                                              'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.GPT.from_pretrained': ( 'Dev_Notes/nanogpt05.html#gpt.from_pretrained',
                                                                                      'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.GPTConfig': ( 'Dev_Notes/nanogpt05.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.MLP': ('Dev_Notes/nanogpt05.html#mlp', 'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.MLP.__init__': ( 'Dev_Notes/nanogpt05.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt5.py'),
                                       'buildNanoGPT.nanogpt5.MLP.forward': ( 'Dev_Notes/nanogpt05.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt5.py')},
            'buildNanoGPT.nanogpt6': { 'buildNanoGPT.nanogpt6.Block': ('Dev_Notes/nanogpt06.html#block', 'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.Block.__init__': ( 'Dev_Notes/nanogpt06.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.Block.forward': ( 'Dev_Notes/nanogpt06.html#block.forward',
                                                                                'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.CausalSelfAttention': ( 'Dev_Notes/nanogpt06.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt06.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt06.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.DataLoaderLite': ( 'Dev_Notes/nanogpt06.html#dataloaderlite',
                                                                                 'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.DataLoaderLite.__init__': ( 'Dev_Notes/nanogpt06.html#dataloaderlite.__init__',
                                                                                          'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.DataLoaderLite.next_batch': ( 'Dev_Notes/nanogpt06.html#dataloaderlite.next_batch',
                                                                                            'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.GPT': ('Dev_Notes/nanogpt06.html#gpt', 'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.GPT.__init__': ( 'Dev_Notes/nanogpt06.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.GPT._init_weights': ( 'Dev_Notes/nanogpt06.html#gpt._init_weights',
                                                                                    'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.GPT.forward': ( 'Dev_Notes/nanogpt06.html#gpt.forward',
                                                                              'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.GPT.from_pretrained': ( 'Dev_Notes/nanogpt06.html#gpt.from_pretrained',
                                                                                      'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.GPTConfig': ( 'Dev_Notes/nanogpt06.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.MLP': ('Dev_Notes/nanogpt06.html#mlp', 'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.MLP.__init__': ( 'Dev_Notes/nanogpt06.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt6.py'),
                                       'buildNanoGPT.nanogpt6.MLP.forward': ( 'Dev_Notes/nanogpt06.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt6.py')},
            'buildNanoGPT.nanogpt7': { 'buildNanoGPT.nanogpt7.Block': ('Dev_Notes/nanogpt07.html#block', 'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.Block.__init__': ( 'Dev_Notes/nanogpt07.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.Block.forward': ( 'Dev_Notes/nanogpt07.html#block.forward',
                                                                                'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.CausalSelfAttention': ( 'Dev_Notes/nanogpt07.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt07.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt07.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.DataLoaderLite': ( 'Dev_Notes/nanogpt07.html#dataloaderlite',
                                                                                 'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.DataLoaderLite.__init__': ( 'Dev_Notes/nanogpt07.html#dataloaderlite.__init__',
                                                                                          'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.DataLoaderLite.next_batch': ( 'Dev_Notes/nanogpt07.html#dataloaderlite.next_batch',
                                                                                            'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.GPT': ('Dev_Notes/nanogpt07.html#gpt', 'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.GPT.__init__': ( 'Dev_Notes/nanogpt07.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.GPT._init_weights': ( 'Dev_Notes/nanogpt07.html#gpt._init_weights',
                                                                                    'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.GPT.configure_optimizers': ( 'Dev_Notes/nanogpt07.html#gpt.configure_optimizers',
                                                                                           'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.GPT.forward': ( 'Dev_Notes/nanogpt07.html#gpt.forward',
                                                                              'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.GPT.from_pretrained': ( 'Dev_Notes/nanogpt07.html#gpt.from_pretrained',
                                                                                      'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.GPTConfig': ( 'Dev_Notes/nanogpt07.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.MLP': ('Dev_Notes/nanogpt07.html#mlp', 'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.MLP.__init__': ( 'Dev_Notes/nanogpt07.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt7.py'),
                                       'buildNanoGPT.nanogpt7.MLP.forward': ( 'Dev_Notes/nanogpt07.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt7.py')},
            'buildNanoGPT.nanogpt8': { 'buildNanoGPT.nanogpt8.Block': ('Dev_Notes/nanogpt08.html#block', 'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.Block.__init__': ( 'Dev_Notes/nanogpt08.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.Block.forward': ( 'Dev_Notes/nanogpt08.html#block.forward',
                                                                                'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.CausalSelfAttention': ( 'Dev_Notes/nanogpt08.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt08.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt08.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.DataLoaderLite': ( 'Dev_Notes/nanogpt08.html#dataloaderlite',
                                                                                 'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.DataLoaderLite.__init__': ( 'Dev_Notes/nanogpt08.html#dataloaderlite.__init__',
                                                                                          'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.DataLoaderLite.next_batch': ( 'Dev_Notes/nanogpt08.html#dataloaderlite.next_batch',
                                                                                            'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.GPT': ('Dev_Notes/nanogpt08.html#gpt', 'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.GPT.__init__': ( 'Dev_Notes/nanogpt08.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.GPT._init_weights': ( 'Dev_Notes/nanogpt08.html#gpt._init_weights',
                                                                                    'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.GPT.configure_optimizers': ( 'Dev_Notes/nanogpt08.html#gpt.configure_optimizers',
                                                                                           'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.GPT.forward': ( 'Dev_Notes/nanogpt08.html#gpt.forward',
                                                                              'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.GPT.from_pretrained': ( 'Dev_Notes/nanogpt08.html#gpt.from_pretrained',
                                                                                      'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.GPTConfig': ( 'Dev_Notes/nanogpt08.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.MLP': ('Dev_Notes/nanogpt08.html#mlp', 'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.MLP.__init__': ( 'Dev_Notes/nanogpt08.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt8.py'),
                                       'buildNanoGPT.nanogpt8.MLP.forward': ( 'Dev_Notes/nanogpt08.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt8.py')},
            'buildNanoGPT.nanogpt9': { 'buildNanoGPT.nanogpt9.Block': ('Dev_Notes/nanogpt09.html#block', 'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.Block.__init__': ( 'Dev_Notes/nanogpt09.html#block.__init__',
                                                                                 'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.Block.forward': ( 'Dev_Notes/nanogpt09.html#block.forward',
                                                                                'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.CausalSelfAttention': ( 'Dev_Notes/nanogpt09.html#causalselfattention',
                                                                                      'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.CausalSelfAttention.__init__': ( 'Dev_Notes/nanogpt09.html#causalselfattention.__init__',
                                                                                               'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.CausalSelfAttention.forward': ( 'Dev_Notes/nanogpt09.html#causalselfattention.forward',
                                                                                              'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.DataLoaderLite': ( 'Dev_Notes/nanogpt09.html#dataloaderlite',
                                                                                 'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.DataLoaderLite.__init__': ( 'Dev_Notes/nanogpt09.html#dataloaderlite.__init__',
                                                                                          'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.DataLoaderLite.next_batch': ( 'Dev_Notes/nanogpt09.html#dataloaderlite.next_batch',
                                                                                            'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.DataLoaderLite.reset': ( 'Dev_Notes/nanogpt09.html#dataloaderlite.reset',
                                                                                       'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.GPT': ('Dev_Notes/nanogpt09.html#gpt', 'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.GPT.__init__': ( 'Dev_Notes/nanogpt09.html#gpt.__init__',
                                                                               'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.GPT._init_weights': ( 'Dev_Notes/nanogpt09.html#gpt._init_weights',
                                                                                    'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.GPT.configure_optimizers': ( 'Dev_Notes/nanogpt09.html#gpt.configure_optimizers',
                                                                                           'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.GPT.forward': ( 'Dev_Notes/nanogpt09.html#gpt.forward',
                                                                              'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.GPT.from_pretrained': ( 'Dev_Notes/nanogpt09.html#gpt.from_pretrained',
                                                                                      'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.GPTConfig': ( 'Dev_Notes/nanogpt09.html#gptconfig',
                                                                            'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.MLP': ('Dev_Notes/nanogpt09.html#mlp', 'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.MLP.__init__': ( 'Dev_Notes/nanogpt09.html#mlp.__init__',
                                                                               'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.MLP.forward': ( 'Dev_Notes/nanogpt09.html#mlp.forward',
                                                                              'buildNanoGPT/nanogpt9.py'),
                                       'buildNanoGPT.nanogpt9.load_tokens': ( 'Dev_Notes/nanogpt09.html#load_tokens',
                                                                              'buildNanoGPT/nanogpt9.py')}}}
